package processView;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.BevelBorder;
import javax.swing.border.Border;

import controls.EquationField;
import controls.EquationTextField;
import controls.UnitComboBox;

import processModel.Formula;
import processModel.Process;

public class EquationPanel extends JPanel 
{
	private Formula equation;
	private List<JComponent> solutionComponents;
	private List<JComponent> formulaComponents;
	private List<EquationField> equationFields;
	private EquationField selectedField;
	private FormulaPanel controller;
	
	private JButton butEvaluate;
	private JButton butRearrange;
	private JLabel labSolution;
	private JLabel labFormula;
	
	public static final Color COLOR_SELECTED_BG = new Color(128, 255, 255);
	public static final Color COLOR_UNSELECTED_BG = new Color(255, 255, 255);
	public static final Color COLOR_TEXT = new Color(0, 0, 0);
	public static final Color COLOR_LABEL = new Color(192, 192, 192);
	public static final int TEXT_FIELD_WIDTH = 30;
	
	public EquationPanel(FormulaPanel controller)
	{
		super();
		formulaComponents = new ArrayList<JComponent>();
		solutionComponents = new ArrayList<JComponent>();
		equationFields = new ArrayList<EquationField>();
		selectedField = null;
		this.controller = controller;
	}
	
	private void initGUI()
	{
		labFormula = new JLabel();
		add(labFormula);
		labFormula.setText("Formula:");
		labFormula.setBounds(0, 0, 42, 20);

		butRearrange = new JButton();
		add(butRearrange);
		butRearrange.setText("Rearrange");
		butRearrange.setBounds(596, 0, 90, 20);
		butRearrange.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				controller.butRearrangeClicked();
			}
		});
		
		labSolution = new JLabel();
		add(labSolution);
		labSolution.setText("Solution:");
		labSolution.setBounds(0, 26, 42, 20);
		
		butEvaluate = new JButton();
		add(butEvaluate);
		butEvaluate.setText("Evaluate");
		butEvaluate.setBounds(596, 25, 90, 20);
		butEvaluate.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent arg0) {
				controller.evaluateEquation();
			}
		});

		if (equation.mustRearrange()) {
			butEvaluate.setVisible(false);
		}
		if (!equation.canRearrange()) {
			butRearrange.setVisible(false);
		}
	}
	
	/**
	 * Construct all components on this EquationPanel
	 */
	public void constructPanelComponents(Formula formula, boolean answerCorrect)
	{
		this.equation = formula;
		clearAllComponents();
		initGUI();
		constructFormulaArea();
		if (!formula.mustRearrange() && answerCorrect) {
			constructSolutionArea();
		}
		else if (!answerCorrect) {
			butEvaluate.setVisible(false);
			labSolution.setVisible(false);
			if (!formula.isRearrangedEquation()) {
				butRearrange.setVisible(false);
			}
		}
		
		// Give focus to the first textbox
		if (equationFields.size() > 0) {
			setSelectedTextField(equationFields.get(0));
			equationFields.get(0).grabFocus();
		}
	}
	
	private void clearAllComponents()
	{
		solutionComponents.clear();
		formulaComponents.clear();
		equationFields.clear();
		selectedField = null;
		this.removeAll();
		this.repaint();
	}
	
	public void setEnabled(boolean enabled)
	{
		for (Component component : getComponents()) {
			component.setEnabled(enabled);
		}
	}
	
	/**
	 * Sets the currently selected text field on this panel
	 */
	public void setSelectedTextField(EquationField newSelectedTextField)
	{
		if (selectedField != null) {
			selectedField.setBackground(COLOR_UNSELECTED_BG);
			if (selectedField.getText().length() == 0) {
				selectedField.setForeground(COLOR_LABEL);
				selectedField.setFieldValue(selectedField.getLabel());
			}
			checkSizeOfTextField(selectedField, TEXT_FIELD_WIDTH);
		}
		
		selectedField = newSelectedTextField;
		if (selectedField != null) {
			if (selectedField.getText().equals(selectedField.getLabel())) {
				selectedField.setFieldValue("");
				selectedField.setForeground(COLOR_TEXT);
			}
			selectedField.setBackground(COLOR_SELECTED_BG);
			selectedField.grabFocus();
		}
	}
	
	/**
	 * Sets the text of the currently selected text field
	 */
	public void setSelectedFieldText(String text)
	{
		int i = equationFields.indexOf(selectedField) + 1;
		// Check if we want the value or the units
		String subtext = "";
		if (i == equationFields.size()) {
			subtext = text.substring(text.indexOf(' ')+1);
		}
		else {
			int index = text.indexOf(' ');
			if (index >= 0) {
				subtext = text.substring(0, index);
			}
			else {
				subtext = text;
			}
		}
		
		if (selectedField != null) {
			selectedField.setFieldValue(subtext);
			if (i < equationFields.size()) {
				setSelectedTextField(equationFields.get(i));
			} else {
				setSelectedTextField(null);
			}
		}
	}
	
	/**
	 * Create the area for displaying the formula to the student
	 */
	private void constructFormulaArea()
	{
		int x = 100;
		int y = 0;
		for (int i = 0; i < equation.equationElements().length; i++)
		{
			String element = equation.equationElements()[i];
			if (i % 2 == 0) {
				JLabel delimiterLabel = createLabel(element, x, y);
				formulaComponents.add(delimiterLabel);
				x += delimiterLabel.getWidth();
			}
			else {
				JLabel symbolLabel = createLabel(element, x, y, TEXT_FIELD_WIDTH);
				formulaComponents.add(symbolLabel);
				x += TEXT_FIELD_WIDTH;
			}
		}
	}
	
	/**
	 * Create a label with the width automatically determined
	 */
	private JLabel createLabel(String text, int x, int y)
	{
		int width = getWidthOfString(text) + 5;
		return createLabel(text, x, y, width);
	}
	
	/**
	 * Creates a label with the specified width
	 */
	private JLabel createLabel(String text, int x, int y, int width)
	{
		JLabel label = new JLabel();
		add(label);
		label.setHorizontalAlignment(JLabel.CENTER);
		label.setVerticalAlignment(JLabel.CENTER);
		label.setText(text);
		label.setBounds(x, y, width, 20);
		return label;
	}
	
	/**
	 * Construct the area for the student to enter their solution.
	 */
	private void constructSolutionArea()
	{
		int x = 100;
		int y = 26;
		for (int i = 0; i < equation.equationElements().length; i++)
		{
			String element = equation.equationElements()[i];
			if (i % 2 == 0) {
				JLabel label = createLabel(element, x, y);
				solutionComponents.add(label);
				x += label.getWidth();
			}
			else {
				EquationTextField textField = createResizingTextField(element, x, y, TEXT_FIELD_WIDTH);
				equationFields.add(textField);
				solutionComponents.add(textField);
				x += TEXT_FIELD_WIDTH;
			}
		}
		// TODO: scrap this
		/*EquationTextField textField = createResizingTextField("Units", x+20, y, 40);
		equationTextFields.add(textField);
		solutionComponents.add(textField);*/
		UnitComboBox unitField = new UnitComboBox(controller.getProcessApplet().getProcess());
		unitField.setBounds(x+20, y, 50, 19);
		add(unitField);
		equationFields.add(unitField);
		solutionComponents.add(unitField);
	}
	
	/**
	 * Creates an automatically resizing text field for this panel
	 * @param label			The text to display on this text field when it is empty
	 * @param x				The x position of the top-left corner
	 * @param y				The y position of the top-left corner
	 * @param textboxWidth	The width of the text field
	 * @return the text field itself
	 */
	private EquationTextField createResizingTextField(String label, int x, int y, final int textfieldWidth)
	{
		final EquationTextField field = new EquationTextField(this, label);
		add(field);
		field.setBounds(x, y, textfieldWidth, 19);
		field.setEditable(false);
		field.setBackground(Color.WHITE);
		
		field.addKeyListener(new KeyAdapter() {
			 public void keyReleased(KeyEvent evt)
			 {
				 checkSizeOfTextField(field, textfieldWidth);
			 }
		});
		return field;
	}
	
	/**
	 * Resizes the text field if necessary.  The size of the text field is determined by its contents.
	 * This will also rearrange other components so they all remain lined up.
	 * @param field
	 */
	public void checkSizeOfTextField(final EquationField field, int textfieldWidth)
	{
		System.out.println("Resizing:  " + field.getText());
		int i = solutionComponents.indexOf(field);;
		int widthIncrease = (int) (getWidthOfString(field.getText()) - field.getWidth()) + 15;
		if (field.getWidth()+widthIncrease < textfieldWidth) {
			widthIncrease = textfieldWidth - field.getWidth();
		}
		field.setSize(field.getWidth()+widthIncrease, field.getHeight());
		if (i < formulaComponents.size()) {
			JComponent formulaComp = formulaComponents.get(i);
			formulaComp.setSize(formulaComp.getWidth()+widthIncrease, formulaComp.getHeight());
			i++;
			while (i < formulaComponents.size()) {
				formulaComp = formulaComponents.get(i);
				formulaComp.setLocation(formulaComp.getX()+widthIncrease, formulaComp.getY());
				JComponent component = solutionComponents.get(i);
				component.setLocation(component.getX()+widthIncrease, component.getY());
				i++;
			}
			// Move the Unit text field
			JComponent component = solutionComponents.get(solutionComponents.size()-1);
			component.setLocation(component.getX()+widthIncrease, component.getY());
		}
	}
	
	/**
	 * Determines the width of a String in pixels for the default font of a JTextField
	 */
	private int getWidthOfString(String string)
	{
		int result = 0;
		if (string.length() > 0) {
			JTextField field = new JTextField();
			TextLayout t = new TextLayout(string, field.getFont(), new FontRenderContext(null, false, false));
			result = (int) t.getBounds().getWidth();
		}
		return result;
	}
	
	/**
	 * Get the String expression from this panel which can then be evaluated with MathEvaluator
	 */
	public String getExpression()
	{
		String result = "";
		for (int i = 0; i < solutionComponents.size()-1; i++)
		{
			JComponent component = solutionComponents.get(i);
			if (component instanceof JLabel) {
				JLabel label = (JLabel)component;
				result += label.getText();
			}
			else if (component instanceof EquationTextField)
			{
				EquationTextField field = (EquationTextField)component;
				result += field.getTextWithoutLabel();
			}
		}
		return result;
	}
	
	/**
	 * Get the String expression from this panel which can then be evaluated with MathEvaluator
	 */
	public String getExpressionWithRounding()
	{
		String result = "";
		for (int i = 0; i < solutionComponents.size()-1; i++)
		{
			JComponent component = solutionComponents.get(i);
			if (component instanceof JLabel) {
				JLabel label = (JLabel)component;
				result += label.getText();
			}
			else if (component instanceof EquationTextField)
			{
				EquationTextField field = (EquationTextField)component;
				result += Process.roundNumber(field.getTextWithoutLabel());
			}
		}
		return result;
	}
	
	/**
	 * Get the String entered into the units text field
	 */
	public String getUnits()
	{
		return ((EquationField)(solutionComponents.get(solutionComponents.size()-1))).getText();
	}
}
